/*
	Copyright (C) 1994,1995	Edward Der-Hua Liu, Hsin-Chu, Taiwan
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include "xi.h"
#include <string.h>
#include <X11/keysym.h>

#define K_REJECT (0)
#define K_ACCEPT (1)

static char *xin_name;
static int screen;
static Colormap colormap;
static XFontStruct     *afont, *cfont;
static XColor fg_color, bg_color;
static long foreground, background;
static char *geomstr="+100-0";
static int win_obscured;
Display *display;
GC      gc,cgc;

Window          main_win;

static Atom wm_del_win;
Atom xcin_atom;

#define MARGIN 0

static XSizeHints sizehints = {
	PMinSize | PMaxSize | PResizeInc | PBaseSize | PWinGravity,
	0, 0, MCOL, MROW,	/* x, y, width and height */
	MCOL, MROW,	/* Min width and height */
	MCOL, MROW,	/* Max width and height */
	MCOL, MROW,	/* Width and height increments */
	{0, 0}, {0, 0}, /* Aspect ratio - not used */
	2 * MARGIN, 2 * MARGIN,	/* base size */
	NorthWestGravity	/* gravity */
};


static int fwidth, fheight;

void p_err(char *fmt,...)
{
  va_list args;
  
  va_start(args, fmt);
  fprintf(stderr,"%s:", xin_name);
  vfprintf(stderr, fmt, args);
  va_end(args);
  fprintf(stderr,"\n");
  exit(-1);
} 

void error(char *fmt,...)
{
  va_list args;
  
  va_start(args, fmt);
  fprintf(stderr,"%s:", xin_name); 
  vfprintf(stderr, fmt, args);
  va_end(args);
  fprintf(stderr,"\n");
}

void pr_ser(char *s, int len)
{
char ttt[3];
int i;

for(i=0;i<len;i++) {
	ser_b5(s,ttt);
	printf("%s",ttt);
	s+=2;
}

}

char *AscFontName="vga12x24";
char *Big5FontName="et24m";

static struct {
u_long fg, bg;
char *fg_str,*bg_str;
} pc_color[]= {
{0,0,"white","SteelBlue"},
{0,0,"yellow","brown"}
};

static void alloc_pc_colors()
{
  XColor color;
  int i;
	for(i=0;i<2;i++) {
	if (!XParseColor(display,colormap,pc_color[i].fg_str, &color))
		p_err("invalid color %s",pc_color[i].fg_str);
	if (!XAllocColor(display,colormap,&color))
		p_err("can't allocate color %s",pc_color[i].fg_str);
		pc_color[i].fg=color.pixel;
  	if (!XParseColor(display,colormap,pc_color[i].bg_str, &color))
		p_err("invalid color %s",pc_color[i].bg_str);
  	if (!XAllocColor(display,colormap,&color))
    		p_err("can't allocate color %s",pc_color[i].bg_str);
  		pc_color[i].bg=color.pixel;
	}
	foreground=pc_color[0].fg;
	background=pc_color[0].bg;
}


static void load_font( char *fnt_name,
                char *cfnt_name)
{
  int x, y, width, height;
  int flags;

  if (!(afont=XLoadQueryFont(display, fnt_name)))
        p_err("Cannot load ascii font %s", fnt_name);
  if (!(cfont=XLoadQueryFont(display, cfnt_name)))
        p_err("Cannot load Chinese font %s", cfnt_name);
  fwidth=sizehints.width_inc = afont->max_bounds.width;
  fheight=sizehints.height_inc = afont->ascent + afont->descent;
  flags = XParseGeometry(geomstr,&x,&y,&width,&height);
  if (flags & WidthValue) {
      sizehints.flags |= USSize;
  }
  if (flags & HeightValue) {
      sizehints.flags |= USSize;
  }
  
  sizehints.width = sizehints.width * sizehints.width_inc +
    sizehints.base_width;
  sizehints.height = sizehints.height * sizehints.height_inc + 
    sizehints.base_height + 1;
  sizehints.max_width=sizehints.min_width=sizehints.width; 
  sizehints.max_height=sizehints.min_height=sizehints.height;
  if (flags & XValue) {
      if (flags & XNegative) {
          x = DisplayWidth(display,screen) + x - sizehints.width - 2;
          sizehints.win_gravity = NorthEastGravity;
        }
      sizehints.x = x;
      sizehints.flags |= USPosition;
    }
  if (flags & YValue) {
      if (flags & YNegative) {
          y = DisplayHeight(display,screen) + y - sizehints.height - 2;
          sizehints.win_gravity = SouthWestGravity;
          if((flags&XValue)&&(flags&XNegative))
            sizehints.win_gravity = SouthEastGravity;
        }
      sizehints.y = y;
      sizehints.flags |= USPosition;
    }
}


#define MW_EVENTS (	KeyPressMask |\
			FocusChangeMask |\
			StructureNotifyMask |\
			ButtonPressMask |\
			ExposureMask    |\
			PropertyChangeMask | \
			VisibilityChangeMask \
		 )

XClassHint class;

static void create_win(int argc, char **argv)
{
  XWMHints wmhints;
  Cursor cursor;
  XTextProperty name;

  main_win = XCreateSimpleWindow(display,DefaultRootWindow(display),
                                 sizehints.x,sizehints.y,
                                 sizehints.width,sizehints.height,
                                 1,foreground,background);
  
  class.res_name = xin_name;
  class.res_class = "xcin";
  wmhints.input = True;
  wmhints.initial_state = NormalState;
    wmhints.flags = InputHint | StateHint;
  
  if (XStringListToTextProperty(&xin_name,1,&name) == 0) {
      error("cannot allocate window name");
      return;
  }
  XSetWMProperties(display,main_win,&name,&name,argv,argc,
                   &sizehints,&wmhints,&class);
  XFree(name.value);

  XSelectInput(display,main_win,MW_EVENTS);
/*  cursor = XCreateFontCursor(display,XC_xterm);
  XRecolorCursor(display,cursor,&fg_color,&bg_color); 
  XDefineCursor(display,main_win,cursor); */
  XMapWindow(display,main_win);
}

static XErrorHandler XInErrorHandler(Display *dpy, XErrorEvent *event)
{
    p_err("Warning, error event receieved\n");
#if	0
    (void) XmuPrintDefaultErrorMessage (dpy, event, stderr);
#endif	0
}

static u_char tch[MROW][MCOL];	/* store text */
static u_char tat[MROW][MCOL];	/* color attribute */
static u_char tby[MROW][MCOL];  /* indicate if it is a chinese ch */

static void ostr(int x, int y, char *str, int len, u_char att)
{
int ix,iy;
u_long fg,bg;

if (!len) len=strlen(str);
memcpy(&tch[y][x], str, len);
memset(&tat[y][x], att, len);
bzero(&tby[y][x], len);
ix=x*fwidth;
iy=y*fheight+afont->ascent+1;
fg=pc_color[att&0xf].fg;
bg=pc_color[att&0xf].bg;
XSetForeground(display, gc, fg);
XSetBackground(display, gc, bg);
XDrawImageString(display, main_win, gc, ix,iy, str, len);
}

static void costr(int x, int y, char *str, int len, u_char att)
{
int ix,iy;
int i;
u_char *bp=&tby[y][x]; 
u_char *cp=&tat[y][x]; 
u_long fg,bg;

if (!len) len=strlen(str);
memcpy(&tch[y][x], str, len);
memset(&tat[y][x], att, len);
for(i=0;i<len;i++) 
   *(bp+i)= ((i)?0x20:0x40);             /* 0:asc  0x20:CH1   0x40:CH2 */

ix=x*fwidth;
iy=y*fheight+cfont->ascent+1;
fg=pc_color[att&0xf].fg;
bg=pc_color[att&0xf].bg;

XSetForeground(display, cgc, fg);
XSetBackground(display, cgc, bg);
XDrawImageString16(display, main_win, cgc, ix,iy, (XChar2b *)str, len>>1);
}

static void restore()
{
int y;
u_char catt,cbyte;
u_char *p,*q;
int stx;
int len;
GC tgc; 
u_long fg,bg;
char tt[80];

for(y=0;y<MROW;y++) {
  p=&tat[y][0];
  q=&tby[y][0];
  stx=0;
  len=0;
  catt=*p;
  cbyte=*q;

  while (stx < MCOL) {
    if (cbyte & 0x60) cbyte|=0x60;
    while (catt==*p && cbyte==(*q&0x60?*q|0x60:*q) && len+stx<MCOL ) {
	len++; p++; q++;
    }
    tgc=(cbyte&0x60)?cgc:gc;
    fg=pc_color[catt&0xf].fg;
    bg=pc_color[catt&0xf].bg;
    XSetForeground(display, tgc, fg);
    XSetBackground(display, tgc, bg);
    if (cbyte & 0x60)
    XDrawImageString16(display, main_win, tgc, stx*fwidth, (y+1)*fheight-1, 
        (XChar2b *) &tch[y][stx], len>>1);
    else
    XDrawImageString(display, main_win, tgc, stx*fwidth, (y+1)*fheight-1,
        &tch[y][stx], len);
    memcpy(tt,&tch[y][stx],len);
    stx+=len;
    tt[len]=0;
    len=0;
    catt=*p;
    cbyte=*q;
  }
}

}

static xputs(int x, int y, u_char *s, u_char att)
{
u_char ch;
int len=strlen(s);
int cl=0;
u_char *cp, *endp, *strp;
char ctyp;
u_char hb,lb;

strp=cp=s;
endp=s+len;
cl=0;
ctyp=0;

while (cp<endp) {
  hb=*cp;
  if ((hb>=0x81&&hb<=0xfe) && cp+1<endp) {
	lb=*(s+1);
	if (lb>=0x40&&lb<=0x7e || lb>=0xa1&&lb<=0xfe) {
		if (ctyp==1) {
			ostr(strp-s+x,y,strp,cl,att);
			cl=0;
			strp=cp;
		}
		cl+=2;
		cp+=2;
		ctyp=2;
		continue;
        }
  }
  if (ctyp==2) {
	costr(strp-s+x,y,strp,cl,att);
	cl=0;
	strp=cp;
  }
  cl++;
  cp++;
  ctyp=1;
}

if (ctyp==1) ostr(strp-s+x,y,strp,cl,att);
if (ctyp==2) costr(strp-s+x,y,strp,cl,att);

}


int cursor_x=0, cursor_y=0;
static u_char cur_att;
void xprintf(char *fmt, ...)
{
  va_list args;
  char tmp[100];
  
  va_start(args, fmt);
  vsprintf(tmp, fmt, args);
  va_end(args);
  xputs(cursor_x, cursor_y, tmp, cur_att);
  cursor_x+=strlen(tmp);
} 

void gotoxy(int x, int y)
{
  cursor_x=x;
  cursor_y=y; 
}
void set_att(u_char att)
{
  cur_att=att;
}

int visual_bell=0;
void bell()
{
if (visual_bell) {
	XClearWindow(display, main_win);
	XFlush(display);
	usleep(100000);
	restore();	
}
else
XBell(display, 100);
}

u_char fullchar[]=
"@ICH]^ϡAСDGFաס֡H"
"IϢТѢҢӢԢբ֢עآ٢ڢۢܢݢޢߢe@fs"
"@ABCaUb  ";

u_char *hfstr[]={"[b]","[]"};
static u_char statstr[]=" i^ơj";
static int targe_inmd_key=1;
int show_keys[9]={10};
int show_keys_n=1;
static int cur_show_keys=0;
static u_char cur_input_ch[2];
char *tabfname[14]={(char *)1,0,0,(char *)1,0,0};
char TabDir[40]=XCIN_DIR;

void show_target(u_char *s)
{
u_char kk[12];
switch (targe_inmd_key) {
	case 10:
	{	u_char t[3];
		ser_b5(s,t);
		sprintf(kk,"%2X%2X      ", (u_int)t[0], (u_int)t[1]);
		break;
	}
	case 11:
	case 3:
		kk[0]=0;
		break;
	default:
		ser2key(targe_inmd_key,s,kk);
}

gotoxy(30,MROW-1); 
if (kk[0]) xprintf("%s",kk);
else xprintf("          ");
}

#define MaxAllowedClients (20)
static int replied=0;

void sendkey(u_char *s)
{
u_char t[3];

ser_b5(s,t);
#if	0
gotoxy(30,0);
xprintf("%s %2X%2X", t, (u_int)t[0], (u_int)t[1]);
#endif	0
show_target(s);
cur_input_ch[0]=s[0];
cur_input_ch[1]=s[1];
write_reply(t,2,K_ACCEPT);
replied=1;
}

void putkey_b5(u_char *s)
{
write_reply(s,2,K_ACCEPT);
replied=1;
}

void putstr(u_char *s)
{
#if	0
	gotoxy(20,0);
{
	int i;
	for(i=0; i<strlen(s); i++)
		printf("%x ", s[i]);
	puts("");
}
	printf("%d|%s|\n",strlen(s), s);
#endif	0
	write_reply(s,strlen(s), K_ACCEPT);
	replied=1;
}

void putfchr(char ch)
{
	char tt[3];
	int ii=(int)(ch-' ')<<1;
	tt[0]=fullchar[ii];
	tt[1]=fullchar[ii+1];
	tt[2]=0;
	putstr(tt);
}

void puthchr(char ch)
{
	char tt[2];
	tt[0]=ch;
	tt[1]=0;
	putstr(tt);
}

static void no_out()
{	write_reply(0,0, K_ACCEPT);
}

static void reject()
{ write_reply(0,0,K_REJECT);
}

#include "state.h"
InmdState inmdstate=
{0,	/* inp state */
 3	/* current in-method */
};

void SwitchCurInMethod(int usenow)
{ 
switch (CurInMethod) {
case 3:
	init_tab_pho(usenow);
	break;
case 6:
	init_tab_pp(usenow);
	break;
case 10:
	init_inter_code(usenow);
	break;
case 11:
	if (tabfname[11]) InitVocBox(usenow);
	else { 
		CurInMethod=3;
		init_tab_pho(usenow);
	}
	break;
default:
	if (tabfname[CurInMethod]) 
		init_tab(CurInMethod, usenow);
	else {
		CurInMethod=3;
		init_tab_pho(usenow);
	}
}
}

void show_halffull()
{
	gotoxy(9,MROW-1);
	xprintf("%s",hfstr[HalfFull]);
}

int ignore_cli_stat=0;

void switch_active_client(InmdState *state)
{
int i;
if (ignore_cli_stat)
	ignore_cli_stat=0;
else
if (state->_CurInMethod)
	memcpy(&inmdstate,state,sizeof(inmdstate));
gotoxy(0,MROW-1);
for(i=0;i<MCOL;i++) xprintf(" ");
gotoxy(0,MROW-1);
if (!EngChi) xprintf("%s",statstr);
show_halffull();
if (EngChi) SwitchCurInMethod(1);
XFlush(display);
}

void next_inp()
{
int i;

for(i=CurInMethod+1; i!=CurInMethod; i=(i+1)%14)
	if (tabfname[i]) {
		if (i) {
			CurInMethod=i;
			inmdstate.kb_state |= EngChiMask;
			SwitchCurInMethod(1);
		} else
		if (EngChi) {
			inmdstate.kb_state &= ~EngChiMask;
			gotoxy(0,MROW-1);
			xprintf("%s",statstr);
			gotoxy(16,MROW-1);	/* clear prev inp */
			xprintf("          ");
		} else 
			continue;
		break;
	}
}

void toggle_halffull()
{
	inmdstate.kb_state^= HalfFullMask;
	show_halffull();
}

void getch(KeySym ch, int state)
{
int count;
int res=0;
	replied=0;
	if (win_obscured) {
	        XRaiseWindow(display, main_win);
		win_obscured=0;
	}
#if	0
	printf("k:%x st:%x\n", ch, state);
#endif	0
	if ((state&(Mod1Mask|Mod5Mask)) && (state&ShiftMask)) {
		res=lookup_key(tolower(ch));
	} else
	if ((ch==' '||ch==0) && (state & ControlMask)) {
		inmdstate.kb_state^=EngChiMask;
		if (!EngChi) {
			gotoxy(0,MROW-1);
			xprintf("%s",statstr);
			gotoxy(16,MROW-1);	/* clear prev inp */
			xprintf("          ");
			ClrSelArea();
		} else
			SwitchCurInMethod(1);
		XFlush(display);
		no_out();
		return;
	} else
	if (ch==' '  && (state & ShiftMask)) {
		toggle_halffull();
		res=1;
		goto return_s;
	} else
	if ( state&ControlMask && (state&(Mod1Mask|Mod5Mask)) ) {
		switch (ch) {
	 	int ttt;
			case '1':
			case '2':
			case '4':
			case '5':
			case '7':
			case '8':
			case '9':
			case '=':
				if (ch=='=') ttt=12;
				else
					ttt=ch-'0';
				if (!tabfname[ttt])
					break;
				inmdstate.kb_state |= EngChiMask;
				CurInMethod=ttt;
				init_tab(CurInMethod,1);
				break;
			case '3':
				CurInMethod=3;
				inmdstate.kb_state |= EngChiMask;
				init_tab_pho();
				break;
			case '6':
				CurInMethod=6;
				inmdstate.kb_state |= EngChiMask;
				init_tab_pp();
				break;
			case '0':
				CurInMethod=10;
				inmdstate.kb_state |= EngChiMask;
				init_inter_code();
				break;
			case '-':
				if (!tabfname[11]) break;
				CurInMethod=11;
				inmdstate.kb_state |= EngChiMask;
				InitVocBox();
				break;
			case 'i':
			case 'I':
				cur_show_keys=(cur_show_keys+1)%show_keys_n;
				targe_inmd_key=show_keys[cur_show_keys];
				if (cur_input_ch[0])
					show_target(cur_input_ch);
				break;
			case 'R':
			case 'r':
				sendkey(cur_input_ch);
				return;
		}
		XFlush(display);
		no_out();
		return;
	} else
	if (EngChi && !(state&(ControlMask|Mod1Mask|Mod5Mask))
	    && ch!=XK_Control_L) {
			int shiftm=state & ShiftMask;
			res = 0;
			if (shiftm && ch!='<' && ch>' ' &&ch<127 &&
				CurInMethod!=6 && ch!='*' && ch!='?' )  {
				char tt[2];
				if (state&LockMask && islower(ch))
					ch-=0x20;
				else
				if (!(state&LockMask) && isupper(ch))
					ch+=0x20;
				if (HalfFull) putfchr(ch);
				else puthchr(ch);
				return;
			}
			switch (CurInMethod) {
			case 3:
				res=feedkey_pho(ch);
				break;
			case 6:
				res=feedkey_pp(ch, state);
				break;
			case 10:
				if (!shiftm) res=feedkey_intcode(ch);
				break;
			case 11:
				res=feedkey_vocbox(ch);
				break;
			default:
				res=feedkey(ch);
				break;
			}
	} else 
#if	0
	if (((ch==XK_Control_L||ch==XK_Control_R) && (state&ShiftMask)) ||
		((ch==XK_Shift_L||ch==XK_Shift_R) && (state&ControlMask)) ) {
#else
	if ((ch==XK_Control_L && state&ShiftMask) ||
		(ch==XK_Shift_L && state&ControlMask) ) {
#endif

		next_inp();
		res=1;	
		goto return_s;
	}
	if ((!EngChi || !res) &&
		!(state&(ControlMask|Mod1Mask|Mod5Mask)) ) {
		u_char tt[3];
		if (HalfFull && ch>=' ' && ch<=0x7e) {
			putfchr(ch);
		}
	}
return_s:
	if (!replied) {
		if (res)
			no_out();
		else
			reject();
	}
	XFlush(display);
}

#define x_cx(x) (x/fwidth)
#define y_cy(y) (y/fheight)

void button_press(XButtonEvent *ev)
{
int cx,cy;
if (ev->button==Button1) {
	cx=x_cx(ev->x);
	cy=y_cy(ev->y);
	if (cy==MROW-1) {
		if (cx>=0 && cx<=8) next_inp();
		else
		if (cx>=9 && cx<=14) toggle_halffull();
		ignore_cli_stat=1;
	} 
}
}

void run_event()
{
  XEvent event;

for(;;) {
  XNextEvent(display,&event);

    switch(event.type) {
    case KeyPress:
      XRaiseWindow(display, main_win);
      break;
    case ClientMessage:
      if (event.xclient.message_type == xcin_atom ) {
		message_dispatch(&event.xclient);
      } else
      if (event.xclient.format == 32 && event.xclient.data.l[0] == wm_del_win)
        	p_err("Quit");
	break;
    case PropertyNotify:
	property_dispatch(&event.xproperty);
	break;
#if	0
    case MappingNotify:
	XRefreshKeyboardMapping(&event.xmapping);
	break;
#endif	0
    case GraphicsExpose:
    case Expose:
	restore();
	break;
    case ButtonPress:
	button_press((XButtonEvent *) &event);
	break;
    case VisibilityNotify:
	{	XVisibilityEvent *eve=(XVisibilityEvent *)&event;
		if (eve->window==main_win &&
		    (eve->state==VisibilityPartiallyObscured||
		     eve->state==VisibilityFullyObscured	) )
				win_obscured=1;
	}
	break;
    default:
/*	printf("Unknow event %d\n",event.type); */
    }
}

}

typedef struct {
	char *argv;
	char *appdef;
	void *res;
	int  type;	/* 1:char*   4:int */
	char *help;
} ARG;

extern char *phokbm_name;
char display_name[16];
extern char tsfname[];
extern char *phrfname;
extern int going_key;

static ARG arg[]={
{"tabdir",0,	TabDir,			21,	"table directory name"},
{"tsfname",0,	tsfname,		21,	"TsIn file name"},
{"inmd",0,	&CurInMethod,	11,	"Input method: 1-11"},
{"phokbm",0,	&phokbm_name,	1,	"phonetic kbd  zo:01 eten:eten"},
{"fg",0,	&pc_color[0].fg_str,	1,	"foreground color"},
{"bg",0,	&pc_color[0].bg_str,	1,	"background color"},
{"infg",0, 	&pc_color[1].fg_str,	1,	"input area fg"},
{"inbg",0, 	&pc_color[1].bg_str,	1,	"input area bg"},
{"geometry",0,	&geomstr,		1,	"position of xcin +100-0"},
{"g",0,		&geomstr,		1,      "position of xcin +100-0"},
{"fn",0,	&AscFontName,		1,	"Ascii font name"},
{"fnb5",0,	&Big5FontName, 		1,	"Big5 font name"},
{"display",0,	&display_name,		1,	"display name"},
{"in1",0,	&tabfname[1],		1,	"in method 1 (eg) cj.tab"},
{"in2",0,	&tabfname[2],		1,	"in method 2"},
{"in4",0,	&tabfname[4],		1,	"in method 4"},
{"in5",0,	&tabfname[5],		1,	"in method 5"},
{"in6",0,	&tabfname[6],		1,	"in method 6"},
{"in7",0,	&tabfname[7],		1,	"in method 7"},
{"in8",0,	&tabfname[8],		1,	"in method 8 (eg) ar30.tab"},
{"in9",0,	&tabfname[9],		1,	"in method 9"},
{"vocbox",0,	&tabfname[11],		1,	"vocabulary box file name"},
{"in12",0,	&tabfname[12],		1,	"in method 12"},
{"going",0,	&going_key,		14,	"going:  0,1"},
{"phrase",0,	&phrfname,		1,	"phrase name"},
{"vb",0, 	&visual_bell,		14,	"visual bell: 0,1"}
};

void get_arg(int argc, char **argv)
{
int num=sizeof(arg)/sizeof(ARG); 
char *pp;
int i,j;

for(i=0;i<num;i++)
if  (pp=XGetDefault(display, xin_name,
			arg[i].appdef ? arg[i].appdef:arg[i].argv)) {
	switch (arg[i].type) {
	case 1: 
		*(char **)arg[i].res=pp; break;
	case 11:
		*(char *)arg[i].res=atoi(pp); break;
	case 14:
		*(int *)arg[i].res=atoi(pp); break;
	case 21:
		strcpy(arg[i].res,pp);
	}
		
}

for (i=1;i<argc;i++) {
	if (argv[i][0]=='-') {
		for(j=0;j<num;j++)
		if (!strcmp(&argv[i][1],arg[j].argv) && i<argc-1) {
			pp=argv[i+1];
			switch (arg[j].type) {
			case 1:
				*(char **)arg[j].res=pp; break;
			case 11:
				*(char *)arg[j].res=atoi(pp); break;
			case 14:
				*(int *)arg[j].res=atoi(pp); break;
			case 21:
				strcpy(arg[j].res,pp);
			}
			i++;
			break;
		}
		if (j==num) {
			char tt[128];
			error("Unknown arg %s", argv[i]);
			error("Valid args are :");
			error("\t\tSee README for details");
			for(j=0;j<num;j++)
				fprintf(stderr,"\t-%s  %s\n", arg[j].argv,
						arg[j].help);
			exit(-1);
		}
	}
}

}

extern void init_msg_channel(); 

main(int argc, char **argv)
{
int i;
XGCValues gcv;
char str[10];
char *pp;
Window twin;
char *display_name;

puts("+------ XCIN version 1.5a ------+");
puts("| by Edward Der-Hua Liu, Taiwan |");
puts("+-------------------------------+");

xin_name="xcin";
if (!(display_name=getenv("DISPLAY")))
        display_name=":0";

for(i=1;i<argc;i++)
     if (!strcmp(argv[i],"-display") && i<argc-1)
     {   display_name=argv[i+1]; break; }

if (!(display = XOpenDisplay(display_name)))
	p_err("Cannot open display %s", display_name);

XSetErrorHandler((XErrorHandler)XInErrorHandler);

get_arg(argc,argv);

if (TabDir[strlen(TabDir)-1]!='/') strcat(TabDir,"/");
if (CurInMethod<=0 || CurInMethod>11 || 
		(!tabfname[CurInMethod]&&CurInMethod!=6))
	CurInMethod=3;
targe_inmd_key=CurInMethod;

colormap = DefaultColormap(display,screen);
screen = DefaultScreen(display);
alloc_pc_colors();

bzero(tat, sizeof(tat));

load_font(AscFontName,Big5FontName);
create_win(argc,argv);

init_msg_channel();

XSetSelectionOwner(display, xcin_atom, main_win, CurrentTime);

set_att(0);
wm_del_win = XInternAtom(display,"WM_DELETE_WINDOW",False);
XSetWMProtocols(display,main_win,&wm_del_win,1);
gcv.foreground = foreground;
gcv.background = background;
gcv.font=afont->fid;
gc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont,&gcv);
gcv.font=cfont->fid;
cgc = XCreateGC(display,main_win,GCForeground|GCBackground|GCFont,&gcv);

SwitchCurInMethod(0);
gotoxy(0,MROW-1);
xprintf("%s",statstr);
show_halffull();
init_phrase();
run_event();
}
